from __future__ import division

import unittest

import numpy as np

from wyrm.processing import lfilter_zi


class TestLFilterZi(unittest.TestCase):

    COEFFS = 10

    def setUp(self):
        self.b, self.a = np.ones(self.COEFFS), np.ones(self.COEFFS)

    def test_lfilter_1d(self):
        """Output has the correct shape for n=1."""
        zi = lfilter_zi(self.b, self.a)
        self.assertEqual(len(zi), self.COEFFS - 1)

    def test_lfilter_nd(self):
        """Output has the correct shape for n>1."""
        zi = lfilter_zi(self.b, self.a, 7)
        self.assertEqual(zi.shape, (self.COEFFS - 1, 7))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""Plotting methods.

This module contains various plotting methods. There are two types of
plotting methods: the Primitives and the Composites. The Primitives are
the most basic and offer simple, single-plot representations. The
Composites are composed of several primitives and offer more complex
representations.

The primitive plots are those whose name begin with ``ax_``, (e.g.
``ax_scalp``).

In order to get more reasonable defaults for colors etc. you can call
the modules :func:`beautify` method::

    from wyrm import plot
    plot.beautify()

.. warning::

    This module needs heavy reworking! We have yet to find a consistent
    way to handle primitive and composite plots, deal with the fact that
    some plots just manipulate axes, while others operate on figures and
    have to decide on which layer of matplotlib we want to deal with
    (i.e. pyplot, artist or even pylab).

    The API of this module will change and you should not rely on any
    method here.

"""


from __future__ import division

import math

import numpy as np
from scipy import interpolate
import matplotlib as mpl
from matplotlib import axes
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Normalize
from matplotlib import pyplot as plt
from matplotlib import ticker
from matplotlib.patches import Rectangle

from wyrm import processing as proc
from wyrm.processing import CHANNEL_10_20
from wyrm.types import Data


# ############# OLD FUNCTIONS ############################################


def plot_channels(dat, ncols=8, chanaxis=-1, otheraxis=-2):
    """Plot all channels for a continuous or epo.

    In case of an epoched Data object, the classwise average is
    calculated, and for each channel the respective classes are plotted.

    Parameters
    ----------
    dat : Data
        continous or epoched Data object
    ncols : int, optional
        the number of colums in the grid. The number of rows is
        calculated depending on ``ncols`` and the number of channels

    """
    # test if epo
    is_epo = False
    if dat.data.ndim == 3:
        is_epo = True
        dat = proc.calculate_classwise_average(dat)
    ax = []
    n_channels = dat.data.shape[chanaxis]
    nrows = int(np.ceil(n_channels / ncols))
    f, ax = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, sharey=True);
    for i, chan in enumerate(dat.axes[chanaxis]):
        a = ax[i // ncols, i % ncols]
        dat.axes[otheraxis], dat.data.take([i], chanaxis)
        if is_epo:
            for j, name in enumerate(dat.class_names):
                cnt = proc.select_classes(dat, [j])
                a.plot(cnt.axes[otheraxis], cnt.data.take([i], chanaxis).squeeze(), label=name)
        else:
            a.plot(dat.axes[otheraxis], dat.data.take([i], chanaxis).squeeze())
        a.set_title(chan)
        a.axvline(x=0, color='black')
        a.axhline(y=0, color='black')
    plt.legend()


def plot_spatio_temporal_r2_values(dat):
    """Calculate the signed r^2 values and plot them in a heatmap.

    Parameters
    ----------
    dat : Data
        epoched data

    """
    r2 = proc.calculate_signed_r_square(dat)
    r2 *= -1
    max = np.max(np.abs(r2))
    plt.imshow(r2.T, aspect='auto', interpolation='None', vmin=-max, vmax=max, cmap='RdBu')
    ax = plt.gca()
    # TODO: sort front-back, left-right
    # use the locators to fine-tune the ticks
    mask = [True if chan.endswith('z') else False for chan in dat.axes[-1]]

    ax.yaxis.set_major_locator(ticker.FixedLocator(np.nonzero(mask)[0]))
    ax.yaxis.set_major_formatter(ticker.IndexFormatter(dat.axes[-1]))

    ax.xaxis.set_major_locator(ticker.MultipleLocator( np.max(dat.axes[-2]) // 100))
    ax.xaxis.set_major_formatter(ticker.IndexFormatter(['%.1f' % i for i in dat.axes[-2]]))

    plt.xlabel('%s [%s]' % (dat.names[-2], dat.units[-2]))
    plt.ylabel('%s [%s]' % (dat.names[-1], dat.units[-1]))
    plt.tight_layout(True)
    plt.colorbar()
    plt.grid(True)


def plot_spectrogram(spectrogram, freqs):
    extent = 0, len(spectrogram), freqs[0], freqs[-1]
    plt.imshow(spectrogram.transpose(),
               aspect='auto',
               origin='lower',
               extent=extent,
               interpolation='none')
    plt.colorbar()
    plt.ylabel('Frequency [Hz]')
    plt.xlabel('Time')


# ############# COMPOSITE PLOTS ##########################################


def plot_timeinterval(data, r_square=None, highlights=None, hcolors=None,
                      legend=True, reg_chans=None, position=None):
    """Plots a simple time interval.

    Plots all channels of either continuous data or the mean of epoched
    data into a single timeinterval plot.

    Parameters
    ----------
    data : wyrm.types.Data
        Data object containing the data to plot.
    r_square : [values], optional
        List containing r_squared values to be plotted beneath the main
        plot (default: None).
    highlights : [[int, int)]
        List of tuples containing the start point (included) and end
        point (excluded) of each area to be highlighted (default: None).
    hcolors : [colors], optional
        A list of colors to use for the highlights areas (default:
        None).
    legend : Boolean, optional
        Flag to switch plotting of the legend on or off (default: True).
    reg_chans : [regular expression], optional
        A list of regular expressions. The plot will be limited to those
        channels matching the regular expressions. (default: None).
    position : [x, y, width, height], optional
        A Rectangle that limits the plot to its boundaries (default:
        None).

    Returns
    -------
    Matplotlib.Axes or (Matplotlib.Axes, Matplotlib.Axes)
        The Matplotlib.Axes corresponding to the plotted timeinterval
        and, if provided, the Axes corresponding to r_squared values.

    Examples
    --------
    Plots all channels contained in data with a legend.

    >>> plot_timeinterval(data)

    Same as above, but without the legend.

    >>> plot_timeinterval(data, legend=False)

    Adds r-square values to the plot.

    >>> plot_timeinterval(data, r_square=[values])

    Adds a highlighted area to the plot.

    >>> plot_timeinterval(data, highlights=[[200, 400]])

    To specify the colors of the highlighted areas use 'hcolors'.

    >>> plot_timeinterval(data, highlights=[[200, 400]], hcolors=['red'])
    """

    dcopy = data.copy()
    rect_ti_solo = [.07, .07, .9, .9]
    rect_ti_r2 = [.07, .12, .9, .85]
    rect_r2 = [.07, .07, .9, .05]

    if position is None:
        plt.figure()
        if r_square is None:
            pos_ti = rect_ti_solo
        else:
            pos_ti = rect_ti_r2
            pos_r2 = rect_r2
    else:
        if r_square is None:
            pos_ti = _transform_rect(position, rect_ti_solo)
        else:
            pos_ti = _transform_rect(position, rect_ti_r2)
            pos_r2 = _transform_rect(position, rect_r2)

    if reg_chans is not None:
        dcopy = proc.select_channels(dcopy, reg_chans)

    # process epoched data into continuous data using the mean
    if len(data.data.shape) > 2:
        dcopy = Data(np.mean(dcopy.data, axis=0), [dcopy.axes[-2], dcopy.axes[-1]],
                     [dcopy.names[-2], dcopy.names[-1]], [dcopy.units[-2], dcopy.units[-1]])

    ax1 = None
    # plotting of the data
    ax0 = _subplot_timeinterval(dcopy, position=pos_ti, epoch=-1, highlights=highlights,
                                hcolors=hcolors, legend=legend)
    ax0.xaxis.labelpad = 0
    if r_square is not None:
        ax1 = _subplot_r_square(r_square, position=pos_r2)
        ax0.tick_params(axis='x', direction='in', pad=30 * pos_ti[3])

    plt.grid(True)

    if r_square is None:
        return ax0
    else:
        return ax0, ax1


def plot_tenten(data, highlights=None, hcolors=None, legend=False, scale=True,
                reg_chans=None):
    """Plots channels on a grid system.

    Iterates over every channel in the data structure. If the
    channelname matches a channel in the tenten-system it will be
    plotted in a grid of rectangles. The grid is structured like the
    tenten-system itself, but in a simplified manner. The rows, in which
    channels appear, are predetermined, the channels are ordered
    automatically within their respective row. Areas to highlight can be
    specified, those areas will be marked with colors in every
    timeinterval plot.

    Parameters
    ----------
    data : wyrm.types.Data
        Data object containing the data to plot.
    highlights : [[int, int)]
        List of tuples containing the start point (included) and end
        point (excluded) of each area to be highlighted (default: None).
    hcolors : [colors], optional
        A list of colors to use for the highlight areas (default: None).
    legend : Boolean, optional
        Flag to switch plotting of the legend on or off (default: True).
    scale : Boolean, optional
        Flag to switch plotting of a scale in the top right corner of
        the grid (default: True)
    reg_chans : [regular expressions]
        A list of regular expressions. The plot will be limited to those
        channels matching the regular expressions.

    Returns
    -------
    [Matplotlib.Axes], Matplotlib.Axes
        Returns the plotted timeinterval axes as a list of
        Matplotlib.Axes and the plotted scale as a single
        Matplotlib.Axes.

    Examples
    --------
    Plotting of all channels within a Data object

    >>> plot_tenten(data)

    Plotting of all channels with a highlighted area

    >>> plot_tenten(data, highlights=[[200, 400]])

    Plotting of all channels beginning with 'A'

    >>> plot_tenten(data, reg_chans=['A.*'])
    """
    dcopy = data.copy()
    # this dictionary determines which y-position corresponds with which row in the grid
    ordering = {4.0: 0,
                3.5: 0,
                3.0: 1,
                2.5: 2,
                2.0: 3,
                1.5: 4,
                1.0: 5,
                0.5: 6,
                0.0: 7,
                -0.5: 8,
                -1.0: 9,
                -1.5: 10,
                -2.0: 11,
                -2.5: 12,
                -2.6: 12,
                -3.0: 13,
                -3.5: 14,
                -4.0: 15,
                -4.5: 15,
                -5.0: 16}

    # all the channels with their x- and y-position
    system = dict(CHANNEL_10_20)

    # create list with 17 empty lists. one for every potential row of channels.
    channel_lists = []
    for i in range(18):
        channel_lists.append([])

    if reg_chans is not None:
        dcopy = proc.select_channels(dcopy, reg_chans)

    # distribute the channels to the lists by their y-position
    count = 0
    for c in dcopy.axes[-1]:
        if c in system:
            # entries in channel_lists: [<channel_name>, <x-position>, <position in Data>]
            channel_lists[ordering[system[c][1]]].append((c, system[c][0], count))
        count += 1

    # sort the lists of channels by their x-position
    for l in channel_lists:
        l.sort(key=lambda c_list: c_list[1])

    # calculate the needed dimensions of the grid
    columns = list(map(len, channel_lists))
    columns = [value for value in columns if value != 0]

    # add another axes to the first row for the scale
    columns[0] += 1

    plt.figure()
    grid = calc_centered_grid(columns, hpad=.01, vpad=.01)

    # axis used for sharing axes between channels
    masterax = None
    ax = []

    row = 0
    k = 0
    scale_ax = 0

    for l in channel_lists:
        if len(l) > 0:
            for i in range(len(l)):
                ax.append(_subplot_timeinterval(dcopy, grid[k], epoch=-1, highlights=highlights, hcolors=hcolors, labels=False,
                                                legend=legend, channel=l[i][2], shareaxis=masterax))
                if masterax is None and len(ax) > 0:
                    masterax = ax[0]

                # hide the axeslabeling
                plt.tick_params(axis='both', which='both', labelbottom='off', labeltop='off', labelleft='off',
                                labelright='off', top='off', right='off')

                # at this moment just to show what's what
                plt.gca().annotate(l[i][0], (0.05, 0.05), xycoords='axes fraction')

                k += 1

                if row == 0 and i == len(l)-1:
                    # this is the last axes in the first row
                    scale_ax = k
                    k += 1

            row += 1

    # plot the scale axes
    xtext = dcopy.axes[0][len(dcopy.axes[0])-1]
    sc = _subplot_scale(str(xtext) + ' ms', "$\mu$V", position=grid[scale_ax])

    return ax, sc


def plot_scalp(v, channels, levels=25, colormap=None, norm=None, ticks=None,
               annotate=True, position=None):
    """Plots the values 'v' for channels 'channels' on a scalp.

    Calculates the interpolation of the values v for the corresponding
    channels 'channels' and plots it as a contour plot on a scalp. The
    degree of gradients as well as the the appearance of the color bar
    can be adjusted.

    Parameters
    ----------
    v : [value]
        List containing the values of the channels.
    channels : [String]
        List containing the channel names.
    levels : int, optional
        The number of automatically created levels in the contour plot
        (default: 25).
    colormap : matplotlib.colors.colormap, optional
        A colormap to define the color transitions (default: a
        blue-white-red colormap).
    norm : matplotlib.colors.norm, optional
        A norm to define the min and max values (default: 'None', values
        from -10 to 10 are assumed).
    ticks : array([ints]), optional
        An array with values to define the ticks on the colorbar
        (default: 'None', 3 ticks at -10, 0 and 10 are displayed).
    annotate : Boolean, optional
        Flag to switch channel annotations on or off (default: True).
    position : [x, y, width, height], optional
        A Rectangle that limits the plot to its boundaries (default:
        None).

    Returns
    -------
    (Matplotlib.Axes, Matplotlib.Axes)
        Returns a pair of Matplotlib.Axes. The first contains the
        plotted scalp, the second the corresponding colorbar.

    Examples
    --------
    Plots the values v for channels 'channels' on a scalp

    >>> plot_scalp(v, channels)

    This plot has finer gradients through increasing the levels to 50.

    >>> plot_scalp(v, channels, levels=50)

    This plot has a norm and ticks from 0 to 10

    >>> n = matplotlib.colors.Normalize(vmin=0, vmax=10, clip=False)
    >>> t = np.linspace(0.0, 10.0, 3, endpoint=True)
    >>> plot_scalp(v, channels, norm=n, ticks=t)
    """
    rect_scalp = [.05, .05, .8, .9]
    rect_colorbar = [.9, .05, .05, .9]

    fig = plt.gcf()

    if position is None:
        pos_scalp = rect_scalp
        pos_colorbar = rect_colorbar
    else:
        pos_scalp = _transform_rect(position, rect_scalp)
        pos_colorbar = _transform_rect(position, rect_colorbar)

    if norm is None:
        vmax = np.abs(v).max()
        vmin = -vmax
        norm = Normalize(vmin, vmax, clip=False)
    if ticks is None:
        ticks = np.linspace(norm.vmin, norm.vmax, 3)

    a = fig.add_axes(pos_scalp)
    ax0 = ax_scalp(v, channels, ax=a, annotate=annotate, vmin=norm.vmin, vmax=norm.vmax,
                   colormap=colormap)
    a = fig.add_axes(pos_colorbar)
    ax1 = ax_colorbar(norm.vmin, norm.vmax, ax=a, ticks=ticks, colormap=colormap,
                      label='')

    return ax0, ax1


def plot_scalp_ti(v, channels, data, interval, scale_ti=.1, levels=25, colormap=None,
                  norm=None, ticks=None, annotate=True, position=None):
    """Plots a scalp with channels on top

    Plots the values v for channels 'channels' on a scalp as a contour
    plot. Additionaly plots the channels in channels_ti as a
    timeinterval on top of the scalp plot. The individual channels are
    placed over their position on the scalp.

    Parameters
    ----------
    v : [value]
        List containing the values of the channels.
    channels : [String]
        List containing the channel names.
    data : wyrm.types.Data
        Data object containing the continuous data for the overlaying
        timeinterval plots.
    interval : [begin, end)
        Tuple of ints to specify the range of the overlaying
        timeinterval plots.
    scale_ti : float, optional
        The percentage to scale the overlaying timeinterval plots
        (default: 0.1).
    levels : int, optional
        The number of automatically created levels in the contour plot
        (default: 25).
    colormap : matplotlib.colors.colormap, optional
        A colormap to define the color transitions (default: a
        blue-white-red colormap).
    norm : matplotlib.colors.norm, optional
        A norm to define the min and max values. If 'None', values from
        -10 to 10 are assumed (default: None).
    ticks : array([ints]), optional
        An array with values to define the ticks on the colorbar
        (default: None, 3  ticks at -10, 0 and 10 are displayed).
    annotate : Boolean, optional
        Flag to switch channel annotations on or off (default: True).
    position : [x, y, width, height], optional
        A Rectangle that limits the plot to its boundaries (default:
        None).

    Returns
    -------
    ((Matplotlib.Axes, Matplotlib.Axes), [Matplotlib.Axes])
        Returns a tuple of first a tuple with the plotted scalp and its
        colorbar, then a list of all on top plotted timeintervals.
    """
    rect_scalp = [.05, .05, .8, .9]
    rect_colorbar = [.9, .05, .05, .9]

    fig = plt.gcf()

    if position is None:
        pos_scalp = rect_scalp
        pos_colorbar = rect_colorbar
    else:
        pos_scalp = _transform_rect(position, rect_scalp)
        pos_colorbar = _transform_rect(position, rect_colorbar)

    if colormap is None:
        colormap = 'RdBu'
    if ticks is None:
        ticks = np.linspace(-10.0, 10.0, 3, endpoint=True)

    a = fig.add_axes(pos_scalp)
    ax0 = ax_scalp(v, channels, ax=a, annotate=annotate)
    a = fig.add_axes(pos_colorbar)
    ax1 = ax_colorbar(-10, 10, ax=a, ticks=ticks)

    # modification of internally used data if a specific intervals is specified
    cdat = data.copy()
    if interval is not None:
        startindex = np.where(cdat.axes[0] == interval[0])[0][0]
        endindex = np.where(cdat.axes[0] == interval[1])[0][0]
        cdat.axes[0] = cdat.axes[0][startindex:endindex]
        cdat.data = cdat.data[startindex:endindex, :]

    tis = []
    for c in cdat.axes[1]:
        points = get_channelpos(c)
        if points is not None:
            channelindex = np.where(cdat.axes[1] == c)[0][0]

            # dirty: these are the x and y limits of the scalp axes
            minx = -1.15
            maxx = 1.15
            miny = -1.10
            maxy = 1.15

            # transformation of karth. to relative coordinates
            xy = (points[0] + (np.abs(minx))) * (1 / (np.abs(minx) + maxx)), \
                 (points[1] + (np.abs(miny))) * (1 / (np.abs(miny) + maxy))

            pos_c = [xy[0] - (scale_ti / 2), xy[1] - (scale_ti / 2), scale_ti, scale_ti]

            # transformation to fit into the scalp part of the plot
            pos_c = _transform_rect(pos_scalp, pos_c)

            tis.append(_subplot_timeinterval(cdat, position=pos_c, epoch=-1, highlights=None, legend=False,
                                             channel=channelindex, shareaxis=None))

        else:
            print('The channel "' + c + '" was not found in the tenten-system.')

    return (ax0, ax1), tis

# ############# TOOLS ####################################################


def set_highlights(highlights, hcolors=None, set_axes=None):
    """Sets highlights in form of vertical boxes to axes.

    Parameters
    ----------
    highlights : [(start, end)]
        List of tuples containing the start point (included) and end
        point (excluded) of each area to be highlighted.
    hcolors : [colors], optional
        A list of colors to use for the highlight areas (e.g. 'b',
        '#eeefff' or [R, G, B] for R, G, B = [0..1]. If left as None the
        colors blue, gree, red, cyan, magenta and yellow are used.
    set_axes : [matplotlib.axes.Axes], optional
        List of axes to highlights (default: None, all axes of the
        current figure will be highlighted).

    Examples
    ---------
    To create two highlighted areas in all axes of the currently active
    figure. The first area from 200ms - 300ms in blue and the second
    area from 500ms - 600ms in green.

    >>> set_highlights([[200, 300], [500, 600]])
    """
    if highlights is not None:

        if set_axes is None:
            set_axes = plt.gcf().axes

        def highlight(start, end, axis, color, alpha):
            axis.axvspan(start, end, edgecolor='w', facecolor=color, alpha=alpha)
            # the edges of the box are at the moment white. transparent edges
            # would be better.

        # create a standard variety of colors, if nothing is specified
        if hcolors is None:
            hcolors = ['b', 'g', 'r', 'c', 'm', 'y']

        # create a colormask containing #spans colors iterating over specified
        # colors or a standard variety
        colormask = []
        for index, span in enumerate(highlights):
            colormask.append(hcolors[index % len(hcolors)])

        # check if highlights is an instance of the Highlight class
        for p in set_axes:
            for idx, span in enumerate(highlights):
                highlight(span[0], span[1], p, colormask[idx], .5)


def calc_centered_grid(cols_list, hpad=.05, vpad=.05):
    """Calculates a centered grid of Rectangles and their positions.

    Parameters
    ----------
    cols_list : [int]
        List of ints. Every entry represents a row with as many channels
        as the value.
    hpad : float, optional
        The amount of horizontal padding (default: 0.05).
    vpad : float, optional
        The amount of vertical padding (default: 0.05).

    Returns
    -------
    [[float, float, float, float]]
        A list of all rectangle positions in the form of [xi, xy, width,
        height] sorted from top left to bottom right.

    Examples
    --------
    Calculates a centered grid with 3 rows of 4, 3 and 2 columns

    >>> calc_centered_grid([4, 3, 2])

    Calculates a centered grid with more padding

    >>> calc_centered_grid([5, 4], hpad=.1, vpad=.75)
    """
    h = (1 - ((len(cols_list) + 1) * vpad)) / len(cols_list)
    w = (1 - ((max(cols_list) + 1) * hpad)) / max(cols_list)
    grid = []
    row = 1
    for l in cols_list:
        yi = 1 - ((row * vpad) + (row * h))
        for i in range(l):
            # calculate margin on both sides
            m = .5 - (((l * w) + ((l - 1) * hpad)) / 2)
            xi = m + (i * hpad) + (i * w)
            grid.append([xi, yi, w, h])
        row += 1
    return grid

# ############# PRIMITIVE PLOTS ##########################################

def _subplot_timeinterval(data, position, epoch, highlights=None, hcolors=None,
                          labels=True, legend=True, channel=None, shareaxis=None):
    """Creates a new axes with a timeinterval plot.

    Creates a matplotlib.axes.Axes within the rectangle specified by
    'position' and fills it with a timeinterval plot defined by the
    channels and values contained in 'data'.

    Parameters
    ----------
    data : wyrm.types.Data
        Data object containing the data to plot.
    position : Rectangle
        The rectangle (x, y, width, height) where the axes will be
        created.
    epoch : int
        The epoch to be plotted. If there are no epochs this has to be
        '-1'.
    highlights : [[int, int)]
        List of tuples containing the start point (included) and end
        point (excluded) of each area to be highlighted (default: None).
    hcolors : [colors], optional
        A list of colors to use for the highlights areas (default:
        None).
    labels : Boolean, optional
        Flag to switch plotting of the usual labels on or off (default:
        True)
    legend : Boolean, optional
        Flag to switch plotting of the legend on or off (default: True).
    channel : int, optional
        This can be used to plot only a single channel. 'channel' has to
        be the index of the desired channel in data.axes[-1] (default:
        None)
    shareaxis : matplotlib.axes.Axes, optional
        An axes to share x- and y-axis with the new axes (default:
        None).

    Returns
    -------
    matplotlib.axes.Axes
    """
    fig = plt.gcf()

    if shareaxis is None:
        ax = fig.add_axes(position)
    else:
        ax = axes.Axes(fig, position, sharex=shareaxis, sharey=shareaxis)
        fig.add_axes(ax)

    # epoch is -1 when there are no epochs
    if epoch == -1:
        if channel is None:
            ax.plot(data.axes[0], data.data)
        else:
            ax.plot(data.axes[0], data.data[:, channel])
    else:
        if channel is None:
            ax.plot(data.axes[len(data.axes) - 2], data.data[epoch])
        else:
            ax.plot(data.axes[len(data.axes) - 2], data.data[epoch, channel])

    # plotting of highlights
    if highlights is not None:
        set_highlights(highlights, hcolors=hcolors, set_axes=[ax])

    # labeling of axes
    if labels:
        ax.set_xlabel(data.units[0])
        ax.set_ylabel("$\mu$V")

    # labeling of channels
    if legend:
        if channel is None:
            ax.legend(data.axes[len(data.axes) - 1])
        else:
            ax.legend([data.axes[len(data.axes) - 1][channel]])

    ax.grid(True)
    return ax


def _subplot_r_square(data, position):
    """Creates a new axes with colored r-sqaure values.

    Parameters
    ----------
    data : [float]
        A list of floats that will be evenly distributed as colored
        tiles.
    position : Rectangle
        The rectangle (x, y, width, height) where the axes will be
        created.

    Returns
    -------
    matplotlib.axes.Axes
    """
    fig = plt.gcf()
    ax = fig.add_axes(position)
    data = np.tile(data, (1, 1))
    ax.imshow(data, aspect='auto', interpolation='none')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    return ax


def _subplot_scale(xvalue, yvalue, position):
    """Creates a new axes with a simple scale.

    Parameters
    ----------
    xvalue : String
        The text to be presented beneath the x-axis.
    yvalue : String
        The text to be presented next to the y-axis.
    position : Rectangle
        The rectangle (x, y, width, height) where the axes will be created.

    Returns
    -------
    matplotlib.axes.Axes
    """
    fig = plt.gcf()
    ax = fig.add_axes(position)
    for item in [fig, ax]:
        item.patch.set_visible(False)
    ax.axis('off')
    ax.add_patch(Rectangle((1, 1), 3, .2, color='black'))
    ax.add_patch(Rectangle((1, 1), .1, 2, color='black'))
    plt.text(1.5, 2, yvalue)
    plt.text(1.5, .25, xvalue)
    ax.set_ylim([0, 4])
    ax.set_xlim([0, 5])
    return ax


def _transform_rect(rect, template):
    """Calculates the position of a relative notated rectangle within
    another rectangle.

    Parameters
    ----------
    rect : Rectangle
        The container rectangle to contain the other reactangle.
    template : Rectangle
        the rectangle to be contained in the other rectangle.
    """
    assert len(rect) == len(template) == 4, "Wrong inputs : [x, y, width, height]"
    x = rect[0] + (template[0] * rect[2])
    y = rect[1] + (template[1] * rect[3])
    w = rect[2] * template[2]
    h = rect[3] * template[3]
    return [x, y, w, h]

###############################################################################
# Primitives
###############################################################################

def ax_scalp(v, channels, ax=None, annotate=False, vmin=None, vmax=None, colormap=None):
    """Draw a scalp plot.

    Draws a scalp plot on an existing axes. The method takes an array of
    values and an array of the corresponding channel names. It matches
    the channel names with an internal list of known channels and their
    positions to project them correctly on the scalp.

    .. warning:: The behaviour for unkown channels is undefined.

    Parameters
    ----------
    v : 1d-array of floats
        The values for the channels
    channels : 1d array of strings
        The corresponding channel names for the values in ``v``
    ax : Axes, optional
        The axes to draw the scalp plot on. If not provided, the
        currently activated axes (i.e. ``gca()``) will be taken
    annotate : Boolean, optional
        Draw the channel names next to the channel markers.
    vmin, vmax : float, optional
        The display limits for the values in ``v``. If the data in ``v``
        contains values between -3..3 and ``vmin`` and ``vmax`` are set
        to -1 and 1, all values smaller than -1 and bigger than 1 will
        appear the same as -1 and 1. If not set, the maximum absolute
        value in ``v`` is taken to calculate both values.
    colormap : matplotlib.colors.colormap, optional
        A colormap to define the color transitions.

    Returns
    -------
    ax : Axes
        the axes on which the plot was drawn

    See Also
    --------
    ax_colorbar

    """
    if ax is None:
        ax = plt.gca()
    # what if we have an unknown channel?
    points = [get_channelpos(c) for c in channels]
    # calculate the interpolation
    x = [i[0] for i in points]
    y = [i[1] for i in points]
    z = v
    # interplolate the in-between values
    xx = np.linspace(min(x), max(x), 500)
    yy = np.linspace(min(y), max(y), 500)
    xx, yy = np.meshgrid(xx, yy)
    f = interpolate.LinearNDInterpolator(list(zip(x, y)), z)
    zz = f(xx, yy)
    # draw the contour map
    ctr = ax.contourf(xx, yy, zz, 20, vmin=vmin, vmax=vmax, cmap=colormap)
    ax.contour(xx, yy, zz, 5, colors="k", vmin=vmin, vmax=vmax, linewidths=.1)
    # paint the head
    ax.add_artist(plt.Circle((0, 0), 1, linestyle='solid', linewidth=2, fill=False))
    # add a nose
    ax.plot([-0.1, 0, 0.1], [1, 1.1, 1], 'k-')
    # add markers at channels positions
    ax.plot(x, y, 'k+')
    # set the axes limits, so the figure is centered on the scalp
    ax.set_ylim([-1.05, 1.15])
    ax.set_xlim([-1.15, 1.15])
    # hide the frame and axes
    # hiding the axes might be too much, as this will also hide the x/y
    # labels :/
    ax.set_frame_on(False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    # draw the channel names
    if annotate:
        for i in zip(channels, list(zip(x, y))):
            ax.annotate(" " + i[0], i[1])
    ax.set_aspect(1)
    plt.sci(ctr)
    return ax

def ax_colorbar(vmin, vmax, ax=None, label=None, ticks=None, colormap=None):
    """Draw a color bar

    Draws a color bar on an existing axes. The range of the colors is
    defined by ``vmin`` and ``vmax``.

    .. note::

        Unlike the colorbar method from matplotlib, this method does not
        automatically create a new axis for the colorbar. It will paint
        in the currently active axis instead, overwriting any existing
        plots in that axis. Make sure to create a new axis for the
        colorbar.

    Parameters
    ----------
    vmin, vmax : float
        The minimum and maximum values for the colorbar.
    ax : Axes, optional
        The axes to draw the scalp plot on. If not provided, the
        currently activated axes (i.e. ``gca()``) will be taken
    label : string, optional
        The label for the colorbar
    ticks : list, optional
        The tick positions
    colormap : matplotlib.colors.colormap, optional
        A colormap to define the color transitions.

    Returns
    -------
    ax : Axes
        the axes on which the plot was drawn
    """
    if ax is None:
        ax = plt.gca()
    ColorbarBase(ax, norm=Normalize(vmin, vmax), label=label, ticks=ticks, cmap=colormap)

###############################################################################
# Utility Functions
###############################################################################

def get_channelpos(channame):
    """Return the x/y position of a channel.

    This method calculates the stereographic projection of a channel
    from ``CHANNEL_10_20``, suitable for a scalp plot.

    Parameters
    ----------
    channame : str
        Name of the channel, the search is case insensitive.

    Returns
    -------
    x, y : float or None
        The projected point on the plane if the point is known,
        otherwise ``None``

    Examples
    --------

    >>> plot.get_channelpos('C2')
    (0.1720792096741632, 0.0)
    >>> # the channels are case insensitive
    >>> plot.get_channelpos('c2')
    (0.1720792096741632, 0.0)
    >>> # lookup for an invalid channel
    >>> plot.get_channelpos('foo')
    None

    """
    channame = channame.lower()
    for i in CHANNEL_10_20:
        if i[0].lower() == channame:
            # convert the 90/4th angular position into x, y, z
            p = i[1]
            ea, eb = p[0] * (90 / 4), p[1] * (90 / 4)
            ea = ea * math.pi / 180
            eb = eb * math.pi / 180
            x = math.sin(ea) * math.cos(eb)
            y = math.sin(eb)
            z = math.cos(ea) * math.cos(eb)
            # Calculate the stereographic projection.
            # Given a unit sphere with radius ``r = 1`` and center at
            # the origin. Project the point ``p = (x, y, z)`` from the
            # sphere's South pole (0, 0, -1) on a plane on the sphere's
            # North pole (0, 0, 1).
            #
            # The formula is:
            #
            # P' = P * (2r / (r + z))
            #
            # We changed the values to move the point of projection
            # further below the south pole
            mu = 1 / (1.3 + z)
            x *= mu
            y *= mu
            return x, y
    return None


def beautify():
    """Set reasonable defaults matplotlib.

    This method replaces matplotlib's default rgb/cmyk colors with the
    colarized colors. It also does:

    * re-orders the default color cycle
    * sets the default linewidth
    * replaces the defaault 'RdBu' cmap
    * sets the default cmap to 'RdBu'

    Examples
    --------

    You can safely call ``beautify`` right after you've imported the
    ``plot`` module.

    >>> from wyrm import plot
    >>> plot.beautify()

    """
    def to_mpl_format(r, g, b):
        """Convert 0..255 t0 0..1."""
        return r / 256, g / 256, b / 256

    # The solarized color palette
    base03  = to_mpl_format(  0,  43,  54)
    base02  = to_mpl_format(  7,  54,  66)
    base01  = to_mpl_format( 88, 110, 117)
    base00  = to_mpl_format(101, 123, 131)
    base0   = to_mpl_format(131, 148, 150)
    base1   = to_mpl_format(147, 161, 161)
    base2   = to_mpl_format(238, 232, 213)
    base3   = to_mpl_format(253, 246, 227)
    yellow  = to_mpl_format(181, 137,   0)
    orange  = to_mpl_format(203,  75,  22)
    red     = to_mpl_format(220,  50,  47)
    magenta = to_mpl_format(211,  54, 130)
    violet  = to_mpl_format(108, 113, 196)
    blue    = to_mpl_format( 38, 139, 210)
    cyan    = to_mpl_format( 42, 161, 152)
    green   = to_mpl_format(133, 153,   0)

    white   = (1, 1, 1)#base3
    black   = base03

    # Tverwrite the default color values with our new ones. Those
    # single-letter colors are used all over the place in matplotlib, so
    # this setting has a huge effect.
    mpl.colors.ColorConverter.colors = {
        'b': blue,
        'c': cyan,
        'g': green,
        'k': black,
        'm': magenta,
        'r': red,
        'w': white,
        'y': yellow
    }

    # Redefine the existing 'RdBu' (Red-Blue) colormap, with our new
    # colors for red and blue
    cdict = {
        'red'  :  ((0., blue[0], blue[0]), (0.5, white[0], white[0]), (1., magenta[0], magenta[0])),
        'green':  ((0., blue[1], blue[1]), (0.5, white[1], white[1]), (1., magenta[1], magenta[1])),
        'blue' :  ((0., blue[2], blue[2]), (0.5, white[2], white[2]), (1., magenta[2], magenta[2]))
    }
    mpl.cm.register_cmap('RdBu', data=cdict)

    # Reorder the default color cycle
    mpl.rcParams['axes.color_cycle'] = ['b', 'm', 'g', 'r', 'c', 'y', 'k']
    # Set linewidth in plots to 2
    mpl.rcParams['lines.linewidth'] = 2
    # Set default cmap
    mpl.rcParams['image.cmap'] = 'RdBu'




from __future__ import division

import unittest

import numpy as np

from wyrm.io import convert_mushu_data


CHANNELS = 5
SAMPLES = 3
FS = 10


class TestConvertMushuData(unittest.TestCase):

    def setUp(self):
        xv, yv = np.meshgrid(list(range(CHANNELS)), list(range(SAMPLES)))
        self.data = xv * 10 + yv
        self.channels = ['ch %i' % i for i in range(CHANNELS)]
        self.time = np.linspace(0, SAMPLES / FS * 1000, SAMPLES, endpoint=False)
        self.markers = [[-10, 'a'], [0, 'b'], [1e100000, 'c']]

    def test_convert_mushu_data(self):
        """Convert mushu data."""
        cnt = convert_mushu_data(self.data, self.markers, FS, self.channels)
        # data
        np.testing.assert_array_equal(cnt.data, self.data)
        self.assertEqual(cnt.data.shape, (SAMPLES, CHANNELS))
        # axes
        timeaxis = np.linspace(0, SAMPLES / FS * 1000, SAMPLES, endpoint=False)
        np.testing.assert_array_equal(cnt.axes[0], timeaxis)
        np.testing.assert_array_equal(cnt.axes[1], self.channels)
        # names and units
        self.assertEqual(cnt.names, ['time', 'channel'])
        self.assertEqual(cnt.units, ['uV', '#'])
        # fs
        self.assertEqual(cnt.fs, FS)
        # markers
        self.assertEqual(cnt.markers, self.markers)

    def test_convert_mushu_data_copy(self):
        """convert_mushu_data should make a copy of its arguments."""
        data = self.data.copy()
        channels = self.channels[:]
        markers = self.markers[:]
        cnt = convert_mushu_data(data, self.markers, FS, channels)
        # data
        data[0, 0] = -1
        np.testing.assert_array_equal(cnt.data, self.data)
        # axes
        channels[0] = 'FOO'
        np.testing.assert_array_equal(cnt.axes[-1], self.channels)
        # markers
        markers[0][0] = markers[0][0] - 1
        self.assertEqual(cnt.markers, self.markers)

if __name__ == '__main__':
    unittest.main()




"""Data type definitions.

This module provides the basic data types for Wyrm, like the
:class:`Data` and :class:`RingBuffer` classes.

"""


from __future__ import division

import copy
import logging

import numpy as np

from wyrm.processing import append_cnt


logging.basicConfig(level=logging.NOTSET)
logger = logging.getLogger(__name__)


class Data(object):
    """Generic, self-describing data container.

    This data structure is very generic on purpose. The goal here was to
    provide something which can fit the various different known and yet
    unknown requirements for BCI algorithms.

    At the core of ``Data`` is its n-dimensional ``.data`` attribute
    which holds the actual data. Along with the data, there is meta
    information about each axis of the data, contained in ``.axes``,
    ``.names``, and ``.units``.

    Most toolbox methods rely on a *convention* how specific data should
    be structured (i.e. they assume that the channels are always in the
    last dimension). You don't have to follow this convention (or
    sometimes it might not even be possible when trying out new things),
    and all methods, provide an optional parameter to tell them on which
    axis they should work on.

    Continuous Data:
        Continuous Data is usually EEG data and consists of a 2d array
        ``[time, channel]``. Whenever you have continuous data, time and
        channel should be the last two dimensions.

    Epoched Data:
        Epoched data can be seen as an array of (non-epoched) data. The
        epoch should always be the first dimension. Most commonly used is
        epoched continuous EEG data which looks like this: ``[class,
        time, channel]``.

    Feature Vector:
        Similar to Epoched Data, with classes in the first dimension.

    :meth:`Data.__eq__` and :meth:`Data.__ne__` functions are provided
    to test for equality of two Data objects (via ``==`` and ``!=``).
    This method only checks for the known attributes and does not
    guaranty correct result if the Data object contains custom
    attributes. It is mainly used in unittests.

    Parameters
    ----------
    data : ndarray
    axes : nlist of 1darrays
    names : nlist of strings
    units : nlist of strings

    Attributes
    ----------
    data : ndarray
        n-dimensional data array if the array is empty
        (i.e. ``data.size == 0``), the ``Data`` object is assumed to be
        empty
    axes : nlist of 1-darrays
        each element of corresponds to a dimension of ``.data`` (i.e.
        the first one in ``.axes`` to the first dimension in ``.data``
        and so on). The 1-dimensional arrays contain the description of
        the data along the appropriate axis in ``.data``. For example if
        ``.data`` contains Continuous Data, then ``.axes[0]`` should be
        an array of timesteps and ``.axes[1]`` an array of channel names
    names : nlist of strings
        the human readable description of each axis, like 'time', or
        'channel'
    units : nlist of strings
        the human readable description of the unit used for the data in
        ``.axes``

    """
    def __init__(self, data, axes, names, units):
        """Initialize a new ``Data`` object.

        Upon initialization we check if ``axes``, ``names``, and
        ``units`` have the same length and if their respective length
        matches the shape of ``data``.

        Raises
        ------
        AssertionError
            if the lengths of the parameters are not correct.

        """
        if data.size == 0:
            pass
        else:
            assert data.ndim == len(axes) == len(names) == len(units)
            assert [len(a) for a in axes] == list(data.shape)
        self.data = data
        self.axes = [np.array(i) for i in axes]
        self.names = names
        self.units = units

    def __eq__(self, other):
        """Test for equality.

        Don't trust this method it only checks for known attributes and
        assumes equality if those are equal. This method is heavily used
        in unittests.

        Parameters
        ----------
        other : Data

        Returns
        -------
        equal : Boolean
            True if ``self`` and ``other`` are equal, False if not.

        """
        # check if both have the same attributes
        if sorted(self.__dict__.keys()) != sorted(other.__dict__.keys()):
            return False
        # .data
        if not np.array_equal(self.data, other.data):
            return False
        # .axes
        if len(self.axes) != len(other.axes):
            return False
        for i in range(len(self.axes)):
            if self.axes[i].shape != other.axes[i].shape:
                return False
            if not (self.axes[i] == other.axes[i]).all():
                return False
        # .names
        if self.names != other.names:
            return False
        # .units
        if self.units != other.units:
            return False
        # optional extra attributes
        if hasattr(self, 'markers') and self.markers != other.markers:
            return False
        if hasattr(self, 'fs') and self.fs != other.fs:
            return False
        # the stuff we care about seems to be equal, this does not mean
        # the rest we didn't check is, but anyways...
        return True

    def __ne__(self, other):
        """Test for inequality.

        If :func:`__eq__` is implemented and :func:`__ne__` is not,
        strange comparisons evaluate to True like:

        >>> d1 == d2 and d1 != d2

        This method just returns the negation of :meth:`__eq__`. So the
        same restrictions of :meth:`__eq__` about its reliability apply.

        Parameters
        ----------
        other : Data

        Returns
        -------
        equal : Boolean
            True if ``self`` and ``other`` are not equal, False
            otherwise.

        """
        return not self.__eq__(other)

    def __nonzero__(self):
        """Return the truth value for the object instance.

        Similar to Python's built in types we return ``False`` if the
        data instance is empty and ``True`` otherwise. Please note that
        we only check for the size of ``.data`` and ignore other
        attributes like ``.markers`` which might not be empty.

        Examples
        --------

        Easy checking if a data object contains data or not:

        >>> if not cnt:
        ...     continue

        is equivalent to:

        >>> if cnt.data.size == 0:
        ...     continue

        Returns
        -------
        nonzero : int
            ``self.data.size``

        """
        return self.data.size

    # This method was added for Python3 compatibility
    def __bool__(self):
        """Return truth value of the object instance.

        This method returns False if the __nonzero__ value is 0 else
        True.

        Returns
        -------
        truth : Bool
            ``False`` if :func:`__nonzero__` was ``0``, else ``True``.

        See Also
        --------
        :func:`__nonzero__`

        """
        return False if self.__nonzero__() == 0 else True

    def __str__(self):
        """Human readable representation for a data object.

        Returns
        -------
        str : str
            a human readable representation of the data object

        """
        data = 'Data: \n%s' % self.data
        axes = 'Axes: \n%s' % self.axes
        names = 'Names: \n%s' % self.names
        units = 'Units: \n%s' % self.units
        return '\n'.join([data, axes, names, units])

    def copy(self, **kwargs):
        """Return a memory efficient deep copy of ``self``.

        It first creates a shallow copy of ``self``, sets the attributes
        in ``kwargs`` if necessary and returns a deep copy of the
        resulting object.

        Parameters
        ----------
        kwargs : dict, optional
            if provided ``copy`` will try to overwrite the name, value
            pairs after the shallow- and before the deep copy. If no
            ``kwargs`` are provided, it will just return the deep copy.

        Returns
        -------
        dat : Data
            a deep copy of ``self``.

        Examples
        --------
        >>> # perform an ordinary deep copy of dat
        >>> dat2 = dat.copy()
        >>> # perform a deep copy but overwrite .axes first
        >>> dat.axes
        ['time', 'channels']
        >>> dat3 = dat.copy(axes=['foo'], ['bar'])
        >>> dat3.axes
        ['foo', 'bar']
        >>> dat.axes
        ['time', 'channel']

        """
        obj = copy.copy(self)
        for name, value in list(kwargs.items()):
            setattr(obj, name, value)
        return copy.deepcopy(obj)


class RingBuffer(object):
    """Circular Buffer implementation.

    This implementation has a guaranteed upper bound for read and write
    operations as well as a constant memory usage, which is the size of
    the maximum length of the buffer in memory.

    Reading and writing will take at most the time it takes to copy a
    continuous chunk of length ``MAXLEN`` in memory. E.g. for the
    extreme case of storing the last 60 seconds of 64bit data, sampled
    with 1kHz and 128 channels (~60MB), reading a full buffer will take
    ~25ms, as well as writing when storing more than than 60 seconds at
    once. Writing will be usually much faster, as one stores usually
    only a few milliseconds of data per run. In that case writing will
    be a fraction of a millisecond.

    Parameters
    ----------
    length_ms : int
        the length of the ring buffer in milliseconds

    Attributes
    ----------
    length_ms : int
        the length of the ring buffer in milliseconds
    length : int
        the length of the ring buffer in samples
    data : ndarray
        the contents of the ring buffer, you should not read or write
        this attribute directly but via the :meth:`RingBuffer.get` and
        :meth:`RingBuffer.append` methods
    markers : array of [int, str]
        the markers belonging to the data currently in the ring buffer
    full : boolean
        indicates if the buffer has at least ``length`` elements stored
    idx : int
        the starting position of the oldest data in the ring buffer

    Examples
    --------

    >>> rb = RingBuffer(length)
    >>> while True:
    ...     rb.append(amp.get_data())
    ...     buffered = rb.get()
    ...     # do something with buffered


    """

    def __init__(self, length_ms):
        """Initialize the Ringbuffer.

        Parameters
        ----------
        length : int
            the length of the ring buffer in milliseconds

        """
        # the maximum length of the ring buffer in ms
        self.length_ms = length_ms
        # the length of the buffer in samples
        self.length = None
        self.data = None
        self.markers = []
        self.axes = []
        self.units = []
        self.names = []
        self.fs = None
        # indicate if the buffer write was wrapped around at least once
        self.full = False
        # the index where to insert new data (= the start of the oldest
        # data)
        self.idx = 0

    def _move_markers(self, markers, steps):
        """Move marker `steps` samples to the left or right.

        This method respects the sampling frequency of the data.

        Parameters
        ----------
        markers : list of (float, str)
        steps : int
            the number of samples to move the markers (a negative value
            moves the indices to the left)

        Returns
        -------
        markers : list of (float, str)

        """
        shift_ms = 1000 / self.fs * steps
        return [[x[0] + shift_ms, x[1]] for x in markers]

    def append(self, dat):
        """Append data to the Ringbuffer, overwriting old data if necessary.

        Parameters
        ----------
        dat : Data
            a continuous data object

        Raises
        ------
        ValueError
            if the [1:]-dimensions (all but the first one) of ``data``
            does not match the ring buffer dimensions

        """
        assert hasattr(dat, 'markers')
        assert hasattr(dat, 'fs')
        data = dat.data.copy()
        markers = dat.markers[:]
        # we have nothing to append
        if len(data) == 0:
            if markers:
                logger.warning('Received Empty Data with markers. Discarding markers.')
                logger.warning(markers)
            return
        # we append the first time, initialize .data with the correct
        # shape
        if self.data is None:
            self.fs = dat.fs
            self.length = self.length_ms / 1000 * self.fs
            if not self.length.is_integer():
                logger.error('Length is not an integer, please check length_ms and fs. Rounding errors will lead to loss of samples.')
            self.length = int(self.length)
            buffershape = list(data.shape)
            buffershape[0] = self.length
            self.data = np.empty(buffershape)
            self.axes = dat.axes[:]
            self.axes[0] = np.linspace(0, 1000 * self.length / self.fs, self.length, endpoint=False)
            self.names = dat.names[:]
            self.units = dat.units[:]
        # incoming data is bigger than the buffer's capacity
        if len(data) > self.length:
            logger.warning('Discarding data that was longer than the ring buffer.')
            surplus = len(data) - self.length
            data = data[surplus:]
            markers = self._move_markers(markers, -surplus)
        # the markers, please be careful when changing it, this is quite
        # tricky:
        # size of the buffer (0..self.length-1)
        size = self.length if self.full else self.idx
        # append the new markers to the end of the existing ones,
        # shifting the new indices by 'size'
        markers = self._move_markers(markers, size)
        self.markers.extend(markers)
        # if we wrapped around, move all elements to the left by the
        # size of the surplus elements
        if size + len(data) > self.length:
            move = self.length - (size + len(data))
            self.markers = self._move_markers(self.markers, move)
        self.markers = [x for x in self.markers if x[0] >= 0]
        # /end of markers
        # we can write without wrapping around the buffer's end
        if self.idx + len(data) < self.length:
            self.data[self.idx:self.idx+len(data)] = data
            self.idx += len(data)
        # we will wrap around the buffer's end
        else:
            self.full = True
            l1 = self.length - self.idx
            l2 = len(data) - l1
            self.data[-l1:] = data[:l1]
            self.data[:l2] = data[l1:]
            self.idx = l2

    def get(self):
        """Get all buffered data.

        The returned data will have *at most* the length of ``length``.

        Returns
        -------
        data : Data
            the full contents of the ring buffer

        """
        # no data has ever been appended to this ring buffer
        if self.data is None:
            data = np.array([])
            axes = []
        # the ringbuffer wrapped around at least once
        elif self.full:
            data = np.concatenate([self.data[self.idx:], self.data[:self.idx]], axis=0)
            axes = self.axes[:]
        # the ring buffer hasn't been filled completely yet
        else:
            data = self.data[:self.idx].copy()
            axes = self.axes[:]
            axes[0] = axes[0][:self.idx]
        d = Data(data=data, axes=axes, names=self.names[:], units=self.units[:])
        d.markers = self.markers[:]
        d.fs = self.fs
        return d


class BlockBuffer(object):
    """A buffer that returns data chunks in multiples of a block length.

    This buffer is a first-in-first-out (FIFO) buffer that returns data
    in multiples of a desired block length. The block length is defined
    in samples.

    Parameters
    ----------
    samples : int, optional
        the desired block length in samples

    Examples
    --------

    >>> bbuffer = BlockBuffer(10)
    >>> ...
    >>> while 1:
    ... cnt = some_aquisition_method()
    ... # How to use the BlockBuffer
    ... bbuffer.append(cnt)
    ... cnt = bbuffer.get()
    ... if not cnt:
    ...     continue
    ... # after here cnt is guaranteed to be in multiples of 10 samples

    """

    def __init__(self, samples=50):
        """Initialize the Block Buffer.

        Parameters
        ----------
        samples : int, optional
            the desired block length in samples

        """
        self.samples = samples
        self.dat = None

    def append(self, dat):
        """Append data to the Block Buffer.

        This method accumulates the incoming data.

        Parameters
        ----------
        dat : Data
            continuous Data object

        """
        if self.dat is None:
            self.dat = dat.copy()
        elif not dat:
            pass
        else:
            self.dat = append_cnt(self.dat, dat)

    def get(self):
        """Pop the contents of the Block Buffer.

        The data returned has a length of multiples of ``samples``. If
        there is a fraction of ``samples`` data more in the buffer, that
        data is kept and future :meth:`append` operations will append
        new data to it.

        Returns
        -------
        dat : Data
            continuous Data object

        """
        if self.dat is None or self.dat.data.shape[0] < self.samples:
            return Data(np.array([]), [], [], [])
        if self.dat.data.shape[0] % self.samples == 0:
            ret = self.dat.copy()
            self.dat = None
            return ret
        else:
            marker_orig = self.dat.markers[:]
            # number of samples to return
            n = (self.dat.data.shape[0] // self.samples) * self.samples
            # first part
            dat1 = self.dat.copy()
            dat1.data = dat1.data[:n]
            dat1.axes[0] = dat1.axes[0][:n]
            # remaining (incomplete) part
            dat2 = self.dat.copy()
            dat2.data = dat2.data[n:]
            dat2.axes[0] = dat2.axes[0][n:]
            # split the markers
            t0 = dat2.axes[0][0]
            dat1.markers = [x for x in self.dat.markers if x[0] < t0]
            dat2.markers = [x for x in self.dat.markers if x[0] >= t0]
            # align the second part to t0
            dat2.axes[0] -= t0
            dat2.markers = [[x[0] - t0, x[1]] for x in dat2.markers]
            self.dat = dat2
            return dat1




from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import swapaxes

class TestSwapaxes(unittest.TestCase):

    def setUp(self):
        raw = np.arange(2000).reshape(-1, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(0, 4000, 400, endpoint=False)
        fs = 100
        marker = [[100, 'foo'], [200, 'bar']]
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])
        self.dat.fs = fs
        self.dat.markers = marker

    def test_swapaxes(self):
        """Swapping axes."""
        new = swapaxes(self.dat, 0, 1)
        self.assertTrue((new.axes[0] == self.dat.axes[1]).all())
        self.assertTrue((new.axes[1] == self.dat.axes[0]).all())
        self.assertEqual(new.names[0], self.dat.names[1])
        self.assertEqual(new.names[1], self.dat.names[0])
        self.assertEqual(new.units[0], self.dat.units[1])
        self.assertEqual(new.units[1], self.dat.units[0])
        self.assertEqual(new.data.shape[::-1], self.dat.data.shape)
        np.testing.assert_array_equal(new.data.swapaxes(0, 1), self.dat.data)

    def test_swapaxes_copy(self):
        """Swapaxes must not modify argument."""
        cpy = self.dat.copy()
        swapaxes(self.dat, 0, 1)
        self.assertEqual(cpy, self.dat)

    def test_swapaxes_twice(self):
        """Swapping the same axes twice must result in original."""
        dat = swapaxes(self.dat, 0, 1)
        dat = swapaxes(dat, 0, 1)
        self.assertEqual(dat, self.dat)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python


import unittest

import numpy as np

from wyrm.types import RingBuffer, Data


def data_factory(data, axes=None, names=None, units=None, markers=None):
    """Helper method to create Data objects."""
    if len(data) == 0:
        axes = names = units = []
    else:
        if axes is None:
            axes = []
            for i in range(data.ndim):
                a = [i * 10 for i in range(data.shape[i])]
                axes.append(a)
        if names is None:
            names = ['name %i' % i for i in range(data.ndim)]
        if units is None:
            units = ['unit %i' % i for i in range(data.ndim)]
    d = Data(data=data, axes=axes, names=names, units=units)
    d.markers = markers if markers is not None else []
    d.fs = 100
    return d


class TestRingBuffer(unittest.TestCase):

    def setUp(self):
        self.rb = RingBuffer(100)

    def test_add_empty(self):
        """Adding an emtpy array must not modify the ringbuffer."""
        # start with emtpy rb
        empty = data_factory(np.array([]))
        dat0 = self.rb.get()
        self.rb.append(empty)
        dat1 = self.rb.get()
        self.assertEqual(dat0, dat1)
        # the same with something in it
        small = data_factory(np.array([0, 1]))
        self.rb.append(small)
        dat0 = self.rb.get()
        self.rb.append(empty)
        dat1 = self.rb.get()
        self.assertEqual(dat0, dat1)

    def test_add(self):
        """Various adding variations must work."""
        d0 = data_factory(np.array([0, 1, 2]))
        d1 = data_factory(np.array([3, 4, 5]))
        d2 = data_factory(np.array([6, 7, 8, 9, 10]))
        d3 = data_factory(np.array([20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]))
        # non-full-add
        self.rb.append(d0)
        data0 = self.rb.get()
        self.assertEqual(data0, d0)
        # non-full-add
        self.rb.append(d1)
        data1 = self.rb.get()
        self.assertEqual(data1, data_factory(np.concatenate([d0.data, d1.data])))
        # overfull-add
        self.rb.append(d2)
        data2 = self.rb.get()
        self.assertEqual(data2, data_factory(np.concatenate([d0.data, d1.data, d2.data])[-10:]))
        self.rb.append(d3)
        data3 = self.rb.get()
        self.assertEqual(data3, data_factory(d3.data[-10:]))
        # test overfull add on empty buffer
        d0 = data_factory(np.arange(11))
        self.rb = RingBuffer(100)
        self.rb.append(d0)
        data0 = self.rb.get()
        self.assertEqual(data0, data_factory(d0.data[-10:]))
        # test full add on full buffer
        self.rb = RingBuffer(100)
        self.rb.append(data_factory(np.arange(10)))
        d0 = data_factory(np.arange(10) + 10)
        self.rb.append(d0)
        data0 = self.rb.get()
        self.assertEqual(data0, d0)

    def test_add_with_mismatching_dimensions(self):
        """Appending data with mismatching dimension must raise a ValueError."""
        d0 = data_factory(np.arange(8).reshape(2, 4))
        d1 = data_factory(np.arange(6).reshape(2, 3))
        self.rb.append(d0)
        with self.assertRaises(ValueError):
            self.rb.append(d1)

    def test_add_with_markers(self):
        # add three elements to empty buffer pluss three markers
        m = [[0, '0'], [20, '2']]
        d3 = data_factory(np.arange(9).reshape(3, 3))
        d3_w_markers = d3.copy(markers=m)
        self.rb.append(d3_w_markers)
        data = self.rb.get()
        self.assertEqual(data.markers, m)
        # move three more elements into the buffer, the markers should
        # stay the same, as we don't overfilled the buffer yet
        self.rb.append(d3)
        data = self.rb.get()
        self.assertEqual(data.markers, m)
        # now we do it again twice and the first one should disappear
        # the second one should move one position back
        self.rb.append(d3)
        self.rb.append(d3)
        data = self.rb.get()
        self.assertEqual(data.markers, [[0, '2']])
        # again, and no markers should be left
        self.rb.append(d3)
        data = self.rb.get()
        self.assertEqual(data.markers, [])
        m = [[0, '0'], [90, '9'], [100, '10']]
        d11_w_markers = data_factory(np.arange(33).reshape(11, 3), markers=m)
        self.rb.append(d11_w_markers)
        data = self.rb.get()
        self.assertEqual(data.markers, [[80, '9'], [90, '10']])
        # test overfull add on empty buffer
        m0 = [[i * 10, i] for i in range(11)]
        d0_w_markers = data_factory(np.arange(33).reshape(11, 3), markers=m0)
        self.rb = RingBuffer(100)
        self.rb.append(d0_w_markers)
        data = self.rb.get()
        self.assertEqual(data.markers, [[x[0]-10, x[1]] for x in m0[-10:]])
        # test full add on full buffer
        self.rb = RingBuffer(100)
        m0 = [[i, i] for i in range(10)]
        self.rb.append(data_factory(np.arange(30).reshape(10, 3), markers=m0))
        m0 = [[x[0], x[1]+10] for x in m0]
        self.rb.append(data_factory(np.arange(30).reshape(10, 3), markers=m0))
        data = self.rb.get()
        self.assertEqual(data.markers, m0)

    def test_appending_until_twice_full_w_1hz(self):
        """Test appending until twice full with 1Hz."""
        d = data_factory(np.array([0]))
        d.fs = 1
        rb = RingBuffer(10000)
        try:
            for i in range(21):
                rb.append(d)
                rb.get()
        except AssertionError:
            self.fail()

    def test_different_lengths(self):
        """Check if setting the length works correctly."""
        d = data_factory(np.array([1]))
        testdata = [
                # ms, fs, elems
                [1000, 1000, 1000],
                [1000, 100, 100],
                [1000, 10, 10],
                [1000, 1, 1],
                [2000, 1000, 2000],
                [2000, 100, 200],
                [2000, 10, 20],
                [2000, 1, 2],
                [1000, 2000, 2000],
                [1000, 200, 200],
                [1000, 20, 20],
                [1000, 2, 2],
                [1000, 1000, 1000],
                [100, 1000, 100],
                [10, 1000, 10],
                [1, 1000, 1],
                 ]
        # 1000ms / 1000hz -> 1000 elements
        for ms, fs, elems in testdata:
            rb = RingBuffer(ms)
            d.fs = fs
            rb.append(d)
            self.assertEqual(rb.length, elems)

if __name__ == '__main__':
    unittest.main()




from __future__ import division

import unittest

import numpy as np

from wyrm.processing import calculate_whitening_matrix, apply_spatial_filter
from wyrm.types import Data


SAMPLES = 1000
CHANS = 3


class TestCalculateWhitentingMatrix(unittest.TestCase):

    def setUp(self):
        data = np.random.randn(SAMPLES, CHANS)
        data[:, 1] += 0.5 * data[:, 0]
        data[:, 2] -= 0.5 * data[:, 0]
        t = np.arange(SAMPLES)
        chans = ['chan{i}'.format(i=i) for i in range(CHANS)]
        self.cnt = Data(data, [t, chans], ['time', 'channels'], ['ms', '#'])

    def test_shape(self):
        """The whitening filter should have the shape: CHANSxCHANS."""
        a = calculate_whitening_matrix(self.cnt)
        self.assertEqual(a.shape, (CHANS, CHANS))

    def test_diagonal(self):
        """The whitened data should have all 1s on the covariance matrix."""
        a = calculate_whitening_matrix(self.cnt)
        dat2 = apply_spatial_filter(self.cnt, a)
        vals = np.diag(np.cov(dat2.data.T))
        np.testing.assert_array_almost_equal(vals, [1. for i in range(len(vals))])

    def test_zeros(self):
        """The whitened data should have all 0s on the non-diagonals of the covariance matrix."""
        a = calculate_whitening_matrix(self.cnt)
        dat2 = apply_spatial_filter(self.cnt, a)
        cov = np.cov(dat2.data.T)
        # substract the diagonals
        cov -= np.diag(np.diag(cov))
        self.assertAlmostEqual(np.sum(cov), 0)

    def test_calculate_whitening_matrix_copy(self):
        """calculate_whitening_matrix must not modify arguments."""
        cpy = self.cnt.copy()
        calculate_whitening_matrix(self.cnt)
        self.assertEqual(self.cnt, cpy)


if __name__ == '__main__':
    unittest.main()




"""Miscellaneous Helper Methods.


"""


import logging
from functools import wraps


logger = logging.getLogger(__name__)


class deprecated(object):
    """Mark a method deprecated

    This method is used internally to mark methods as deprecated.
    Deprecated methods will print a warning when used but will otherwise
    function as usual.

    Parameters
    ----------
    since : str
        The version number that introduced the deprecation of the
        method.
    alternative : str, optional
        The method that should be used instead.

    """
    def __init__(self, since, alternative=None):
        self.since = since
        self.alternative = alternative

    def __call__(self, f):
        msg = "{f} is deprecated since version {v}.".format(f=f.__name__, v=self.since)
        if self.alternative is not None:
            msg2 = " Please use {f} instead.".format(f=self.alternative)
            msg = msg + msg2
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            logger.warn(msg)
            return f(*args, **kwargs)
        return wrapped_f




#!/usr/bin/env python

from __future__ import division

import unittest

import numpy as np
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

from wyrm.types import Data
from wyrm import plot
from wyrm.processing import CHANNEL_10_20


class TestPlot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        x = np.linspace(0, 1000, 1000, endpoint=False)
        y = 5 * np.sin(np.pi * x / 300)
        data = np.tile(y[:, None], (1, len(CHANNEL_10_20)))
        data += np.random.normal(size=data.shape)
        axes = [x, [i[0] for i in CHANNEL_10_20]]
        cls.cnt = Data(data=data, axes=axes, names=['time', 'channel'], units=['ms', '#'])
        cls.cnt.fs = 1000

        classes = [0, 1] * 5
        data = np.array([data]*10)
        data[::2] *= 0.5
        axes = [classes, x, [i[0] for i in CHANNEL_10_20]]
        cls.epo = Data(data=data, axes=axes, names=['class', 'time', 'channel'], units=['#', 'ms', '#'])
        cls.epo.fs = 1000
        cls.epo.class_names = ['class 1', 'class 2']

        plot.beautify()

    def setUp(self):
        plt.figure(figsize=(16, 9))

    def tearDown(self):
        test_method = self._testMethodName
        plt.suptitle(test_method)
        plt.savefig(test_method + '.jpg')
        plt.close()

###############################################################################

    def test_plot_channels_cnt(self):
        plot.plot_channels(self.cnt)

    def test_plot_channels_epo(self):
        plot.plot_channels(self.epo)

    def test_plot_spatio_temporal_r2_values(self):
        plot.plot_spatio_temporal_r2_values(self.epo)

    def test_plot_spectrogram(self):
        plot.plot_spectrogram(np.random.random((10, 100)), np.arange(100))

###############################################################################

    def test_plot_timeinterval(self):
        _, ax = plt.subplots(2, 1)
        plt.sca(ax[0])
        plot.plot_timeinterval(self.cnt, np.random.random(1000), [[200, 300], [500, 700]])
        plt.sca(ax[1])
        plot.plot_timeinterval(self.cnt)

    def test_plot_tenten(self):
        plot.plot_tenten(self.cnt)

    def test_plot_scalp(self):
        plot.plot_scalp(self.cnt.data[0, :], self.cnt.axes[-1])

    def test_plot_scalp_ti(self):
        plot.plot_scalp_ti(self.cnt.data[0, :], self.cnt.axes[-1], self.cnt, [100, 700])


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import rereference
from wyrm.processing import swapaxes


CHANS = 5
SAMPLES = 20
EPOS = 3


class TestRereference(unittest.TestCase):

    def setUp(self):
        dat = np.zeros((SAMPLES, CHANS))
        # [-10, -9, ... 20)
        dat[:, 0] = np.arange(SAMPLES) - SAMPLES/2
        channels = ['chan{i}'.format(i=i) for i in range(CHANS)]
        time = np.arange(SAMPLES)
        self.cnt = Data(dat, [time, channels], ['time', 'channels'], ['ms', '#'])
        # construct epo
        epo_dat = np.array([dat + i for i in range(EPOS)])
        classes = ['class{i}'.format(i=i) for i in range(EPOS)]
        self.epo = Data(epo_dat, [classes, time, channels], ['class', 'time', 'channels'], ['#', 'ms', '#'])

    def test_rereference_cnt(self):
        """Rereference channels (cnt)."""
        cnt_r = rereference(self.cnt, 'chan0')
        dat_r = np.linspace(SAMPLES/2, -SAMPLES/2, SAMPLES, endpoint=False)
        dat_r = [dat_r for i in range(CHANS)]
        dat_r = np.array(dat_r).T
        dat_r[:, 0] = 0
        np.testing.assert_array_equal(cnt_r.data, dat_r)

    def test_rereference_epo(self):
        """Rereference channels (epo)."""
        epo_r = rereference(self.epo, 'chan0')
        dat_r = np.linspace(SAMPLES/2, -SAMPLES/2, SAMPLES, endpoint=False)
        dat_r = [dat_r for i in range(CHANS)]
        dat_r = np.array(dat_r).T
        dat_r[:, 0] = 0
        dat_r = np.array([dat_r for i in range(EPOS)])
        np.testing.assert_array_equal(epo_r.data, dat_r)

    def test_raise_value_error(self):
        """Raise ValueError if channel not found."""
        with self.assertRaises(ValueError):
            rereference(self.cnt, 'foo')

    def test_case_insensitivity(self):
        """rereference should not care about case."""
        try:
            rereference(self.cnt, 'ChAN0')
        except ValueError:
            self.fail()

    def test_rereference_copy(self):
        """rereference must not modify arguments."""
        cpy = self.cnt.copy()
        rereference(self.cnt, 'chan0')
        self.assertEqual(self.cnt, cpy)

    def test_rereference_swapaxes(self):
        """rereference must work with nonstandard chanaxis."""
        dat = rereference(swapaxes(self.epo, 1, 2), 'chan0', chanaxis=1)
        dat = swapaxes(dat, 1, 2)
        dat2 = rereference(self.epo, 'chan0')
        self.assertEqual(dat, dat2)


if __name__ == '__main__':
    unittest.main()




from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import correct_for_baseline
from wyrm.processing import swapaxes

class TestCorrectForBaseline(unittest.TestCase):

    def setUp(self):
        ones = np.ones((10, 5))
        # three blocks: 1s, -1s, and 0s
        cnt_data = np.concatenate([ones, -ones, 0*ones], axis=0)
        classes = [0, 0, 0]
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.linspace(-1000, 2000, 30, endpoint=False)
        epo_data = np.array([cnt_data, cnt_data, cnt_data])
        self.dat = Data(epo_data, [classes, time, channels], ['class', 'time', 'channels'], ['#', 'ms', '#'])

    def test_correct_for_baseline_epo(self):
        """Test baselineing w/ epo like."""
        dat2 = correct_for_baseline(self.dat, [-1000, 0])
        np.testing.assert_array_equal(dat2.data, self.dat.data - 1)
        dat2 = correct_for_baseline(self.dat, [0, 1000])
        np.testing.assert_array_equal(dat2.data, self.dat.data + 1)
        dat2 = correct_for_baseline(self.dat, [1000, 2000])
        np.testing.assert_array_equal(dat2.data, self.dat.data)
        # the full dat interval
        dat = correct_for_baseline(self.dat, [self.dat.axes[-2][0], self.dat.axes[-2][-1]])
        np.testing.assert_array_equal(self.dat.data, dat.data)

    def test_correct_for_baseline_cnt(self):
        """Test baselineing w/ cnt like."""
        data = self.dat.data[0]
        axes = [np.linspace(-1000, 2000, 30, endpoint=False), self.dat.axes[-1]]
        units = self.dat.units[1:]
        names = self.dat.names[1:]
        dat = self.dat.copy(data=data, axes=axes, names=names, units=units)
        dat2 = correct_for_baseline(dat, [-1000, 0])
        np.testing.assert_array_equal(dat2.data, dat.data - 1)
        dat2 = correct_for_baseline(dat, [0, 1000])
        np.testing.assert_array_equal(dat2.data, dat.data + 1)
        dat2 = correct_for_baseline(dat, [1000, 2000])
        np.testing.assert_array_equal(dat2.data, dat.data)
        # the full interval
        dat2 = correct_for_baseline(dat, [dat.axes[-2][0], dat.axes[-2][-1]])
        np.testing.assert_array_equal(dat2.data, dat.data)

    def test_ival_checks(self):
        """Test for malformed ival parameter."""
        with self.assertRaises(AssertionError):
            correct_for_baseline(self.dat, [0, -1])
        with self.assertRaises(AssertionError):
            correct_for_baseline(self.dat, [self.dat.axes[-2][0]-1, 0])
        with self.assertRaises(AssertionError):
            correct_for_baseline(self.dat, [0, self.dat.axes[-2][1]+1])

    def test_correct_for_baseline_copy(self):
        """Correct for baseline must not modify dat argument."""
        cpy = self.dat.copy()
        correct_for_baseline(self.dat, [-1000, 0])
        self.assertEqual(cpy, self.dat)

    def test_correct_for_baseline_swapaxes(self):
        """Correct for baseline must work with nonstandard timeaxis."""
        dat = correct_for_baseline(swapaxes(self.dat, 0, 1), [-1000, 0], timeaxis=0)
        dat = swapaxes(dat, 0, 1)
        dat2 = correct_for_baseline(self.dat, [-1000, 0])
        self.assertEqual(dat, dat2)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import select_channels
from wyrm.processing import swapaxes

class TestSelectChannels(unittest.TestCase):

    def setUp(self):
        raw = np.arange(20).reshape(4, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.arange(4)
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])

    def test_select_channels(self):
        """Selecting channels with an array of regexes."""
        channels = self.dat.data.copy()
        self.dat = select_channels(self.dat, ['ca.*', 'cc1'])
        np.testing.assert_array_equal(self.dat.axes[-1],  np.array(['ca1', 'ca2', 'cc1']))
        np.testing.assert_array_equal(self.dat.data, channels[:, np.array([0, 1, -1])])

    def test_select_channels_inverse(self):
        """Removing channels with an array of regexes."""
        channels = self.dat.data.copy()
        self.dat = select_channels(self.dat, ['ca.*', 'cc1'], invert=True)
        np.testing.assert_array_equal(self.dat.axes[-1],  np.array(['cb1', 'cb2']))
        np.testing.assert_array_equal(self.dat.data, channels[:, np.array([2, 3])])

    def test_select_channels_copy(self):
        """Select channels must not change the original parameter."""
        cpy = self.dat.copy()
        select_channels(self.dat, ['ca.*'])
        self.assertEqual(cpy, self.dat)

    def test_select_channels_swapaxis(self):
        """Select channels works with non default chanaxis."""
        dat1 = select_channels(swapaxes(self.dat, 0, 1), ['ca.*'], chanaxis=0)
        dat1 = swapaxes(dat1, 0, 1)
        dat2 = select_channels(self.dat, ['ca.*'])
        self.assertEqual(dat1, dat2)

if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np
np.random.seed(42)

from wyrm.types import Data
from wyrm.processing import apply_csp


class TestApplyCSP(unittest.TestCase):

    EPOCHS = 50
    SAMPLES = 100
    SOURCES = 2
    CHANNELS = 10

    def setUp(self):
        # create a random noise signal with 50 epochs, 100 samples, and
        # 2 sources
        # even epochs and source 0: *= 5
        # odd epochs and source 1: *= 5
        self.s = np.random.randn(self.EPOCHS, self.SAMPLES, self.SOURCES)
        self.s[ ::2, :, 0] *= 5
        self.s[1::2, :, 1] *= 5
        # the mixmatrix which converts our sources to channels
        # X = As + noise
        self.A = np.random.randn(self.CHANNELS, self.SOURCES)
        # our 'signal' which 50 epochs, 100 samples and 10 channels
        self.X = np.empty((self.EPOCHS, self.SAMPLES, self.CHANNELS))
        for i in range(self.EPOCHS):
            self.X[i] = np.dot(self.A, self.s[i].T).T
        noise = np.random.randn(self.EPOCHS, self.SAMPLES, self.CHANNELS) * 0.01
        self.X += noise

        a = np.array([1 for i in range(self.X.shape[0])])
        a[0::2] = 0
        axes = [a, np.arange(self.X.shape[1]), np.arange(self.X.shape[2])]
        self.epo = Data(self.X, axes=axes, names=['class', 'time', 'channel'], units=['#', 'ms', '#'])
        self.epo.class_names = ['foo', 'bar']

        self.filter = np.random.random((self.CHANNELS, self.CHANNELS))

    def test_apply_csp(self):
        """apply_csp."""
        dat = apply_csp(self.epo, self.filter)
        # reduce the channels down to 2, the rest of the shape should
        # stay the same
        self.assertEqual(self.epo.data.shape[0], dat.data.shape[0])
        self.assertEqual(self.epo.data.shape[1], dat.data.shape[1])
        self.assertEqual(2, dat.data.shape[2])
        # new name for csp axis
        self.assertEqual(dat.names[-1], 'CSP Channel')
        # check if the dot product was calculated correctly
        d = np.array([np.dot(self.epo.data[i], self.filter[:, [0, -1]]) for i in range(self.epo.data.shape[0])])
        np.testing.assert_array_equal(d, dat.data)


    #def test_apply_csp_swapaxes(self):
    #    """apply_csp must work with nonstandard classaxis."""
    #    dat = apply_csp(swapaxes(self.epo, 0, 1), self.filter.T, classaxis=1)
    #    #dat = swapaxes(dat, 0, 1)
    #    print dat.data.shape
    #    dat2 = apply_csp(self.epo, self.filter)
    #    print
    #    print dat2.data.shape
    #    self.assertEqual(dat, dat2)

    def test_apply_csp_copy(self):
        """apply_csp must not modify argument."""
        cpy = self.epo.copy()
        apply_csp(self.epo, self.filter)
        self.assertEqual(self.epo, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np
np.random.seed(42)

from wyrm.types import Data
from wyrm.processing import calculate_spoc


class TestCalculateSpoc(unittest.TestCase):

    EPOCHS = 50
    SAMPLES = 100
    SOURCES = 2
    CHANNELS = 10

    def setUp(self):
        # generate sources with independent variance modulations, the
        # first source will be the target source
        z = np.abs(np.random.randn(self.EPOCHS, self.SOURCES))
        for i in range(self.SOURCES):
            z[:, i] /= z[:, i].std()
        self.s = np.random.randn(self.EPOCHS, self.SAMPLES, self.SOURCES)
        for i in range(self.SOURCES):
            for j in range(self.EPOCHS):
                self.s[j, :, i] *= z[j, i]
        # the mixmatrix which converts our sources to channels
        # X = As + noise
        self.A = np.random.randn(self.CHANNELS, self.SOURCES)
        # our 'signal' which 50 epochs, 100 samples and 10 channels
        self.X = np.empty((self.EPOCHS, self.SAMPLES, self.CHANNELS))
        for i in range(self.EPOCHS):
            self.X[i] = np.dot(self.A, self.s[i].T).T
        noise = np.random.randn(self.EPOCHS, self.SAMPLES, self.CHANNELS) * 0.01
        self.X += noise
        # convert to epo
        axes = [z[:, 0], np.arange(self.X.shape[1]), np.arange(self.X.shape[2])]
        self.epo = Data(self.X,
                axes=axes,
                names=['target_variable', 'time', 'channel'],
                units=['#', 'ms', '#'])

    def test_d(self):
        """Test if the list of lambdas is reverse-sorted and the first one > 0."""
        W, A_est, d = calculate_spoc(self.epo)
        self.assertTrue(d[0] > 0)
        self.assertTrue(np.all(d == np.sort(d)[::-1]))

    def test_A(self):
        """Test if A_est is elementwise almost equal A."""
        W, A_est, d = calculate_spoc(self.epo)
        # A and A_est can have a different scaling, after normalizing
        # and correcting for sign, the first pattern should be almost
        # equal the source pattern
        idx = np.argmax(np.abs(A_est[:, 0]))
        A_est[:, 0] /= A_est[idx, 0]
        idx = np.argmax(np.abs(self.A[:, 0]))
        self.A[:, 0] /= self.A[idx, 0]
        # check elementwise if A[:, 0] almost A_est[:, 0]
        epsilon = 0.01
        diff = self.A[:, 0] - A_est[:, 0]
        diff = np.abs(diff)
        diff = np.sum(diff) / self.A.shape[0]
        self.assertTrue(diff < epsilon)

    def test_s(self):
        """Test if s_est is elementwise almost equal s."""
        W, A_est, d = calculate_spoc(self.epo)
        # applying the filter to X gives us s_est which should be almost
        # equal s
        s_est = np.empty(self.s.shape[:2])
        for i in range(self.EPOCHS):
            s_est[i] = np.dot(self.X[i], W[:, 0])
        s_true = self.s[..., 0]
        epsilon = 0.001

        # correct for scale
        s_true /= s_true.std()
        s_est /= s_est.std()

        # correct for sign
        s_true = np.abs(s_true)
        s_est = np.abs(s_est)

        diff = np.sum(s_true - s_est) / (self.s.shape[0] * self.s.shape[1])
        self.assertTrue(diff < epsilon)

    #def test_calculate_signed_r_square_swapaxes(self):
    #    """caluclate_r_square must work with nonstandard classaxis."""
    #    dat = calculate_signed_r_square(swapaxes(self.dat, 0, 2), classaxis=2)
    #    # the class-axis just dissapears during
    #    # calculate_signed_r_square, so axis 2 becomes axis 1
    #    dat = dat.swapaxes(0, 1)
    #    dat2 = calculate_signed_r_square(self.dat)
    #    np.testing.assert_array_equal(dat, dat2)

    def test_calculate_signed_r_square_copy(self):
        """caluclate_r_square must not modify argument."""
        cpy = self.epo.copy()
        calculate_spoc(self.epo)
        self.assertEqual(self.epo, cpy)


if __name__ == '__main__':
    unittest.main()



from __future__ import division

import unittest

import numpy as np

from wyrm.processing import apply_spatial_filter
from wyrm.processing import swapaxes
from wyrm.types import Data


SAMPLES = 1000
CHANS = 3
EPOS = 5


class TestApplySpatialFilter(unittest.TestCase):

    def setUp(self):
        data = np.random.randn(SAMPLES, CHANS)
        data[:, 1] += 0.5 * data[:, 0]
        data[:, 2] -= 0.5 * data[:, 0]
        t = np.arange(SAMPLES)
        chans = ['chan{i}'.format(i=i) for i in range(CHANS)]
        self.cnt = Data(data, [t, chans], ['time', 'channels'], ['ms', '#'])

        # construct epo
        epo_dat = np.array([data for i in range(EPOS)])
        classes = ['class{i}'.format(i=i) for i in range(EPOS)]
        self.epo = Data(epo_dat, [classes, t, chans], ['class', 'time', 'channels'], ['#', 'ms', '#'])

        # my little spatial filter
        self.w = np.array([[ 0, 0.5, 1],
                           [-1, 0.5, 0],
                           [ 1, 0.5, 0]])


    def test_shape(self):
        """The spatial filtered data should keep its shape."""
        cnt_f = apply_spatial_filter(self.cnt, self.w)
        epo_f = apply_spatial_filter(self.epo, self.w)
        self.assertEqual(self.cnt.data.shape, cnt_f.data.shape)
        self.assertEqual(self.epo.data.shape, epo_f.data.shape)


    def test_spatial_filter_cnt(self):
        """Spatial filtering should work with cnt."""
        cnt_f = apply_spatial_filter(self.cnt, self.w)
        # chan 0
        np.testing.assert_array_equal(cnt_f.data[:, 0], self.cnt.data[:, 2] - self.cnt.data[:, 1])
        # chan 1
        np.testing.assert_array_equal(cnt_f.data[:, 1], 0.5 * np.sum(self.cnt.data, axis=-1))
        # chan 2
        np.testing.assert_array_equal(cnt_f.data[:, 2], self.cnt.data[:, 0])


    def test_spatial_filter_epo(self):
        """Spatial filtering should work with epo."""
        cnt_f = apply_spatial_filter(self.cnt, self.w)
        epo_f = apply_spatial_filter(self.epo, self.w)
        for i in range(EPOS):
            np.testing.assert_array_equal(epo_f.data[i, ...], cnt_f.data)


    def test_prefix_and_postfix(self):
        """Prefix and Postfix are mutual exclusive."""
        with self.assertRaises(ValueError):
            apply_spatial_filter(self.cnt, self.w, prefix='foo', postfix='bar')


    def test_prefix(self):
        """Apply prefix correctly."""
        cnt_f = apply_spatial_filter(self.cnt, self.w, prefix='foo')
        self.assertEqual(cnt_f.axes[-1], ['foo'+str(i) for i in range(CHANS)])


    def test_prefix_w_wrong_type(self):
        """Raise TypeError if prefix is neither None or str."""
        with self.assertRaises(TypeError):
            apply_spatial_filter(self.cnt, self.w, prefix=1)


    def test_postfix(self):
        """Apply postfix correctly."""
        cnt_f = apply_spatial_filter(self.cnt, self.w, postfix='foo')
        self.assertEqual(cnt_f.axes[-1], [c+'foo' for c in self.cnt.axes[-1]])


    def test_postfix_w_wrong_type(self):
        """Raise TypeError if postfix is neither None or str."""
        with self.assertRaises(TypeError):
            apply_spatial_filter(self.cnt, self.w, postfix=1)


    def test_apply_spatial_filter_copy(self):
        """apply_spatial_filter must not modify arguments."""
        cpy = self.cnt.copy()
        apply_spatial_filter(self.cnt, self.w)
        self.assertEqual(self.cnt, cpy)


    def test_apply_spatial_filter_swapaxes(self):
        """apply_spatial_filter must work with nonstandard chanaxis."""
        epo_f = apply_spatial_filter(swapaxes(self.epo, 1, -1), self.w,  chanaxis=1)
        epo_f = swapaxes(epo_f, 1, -1)
        epo_f2 = apply_spatial_filter(self.epo, self.w)
        self.assertEqual(epo_f, epo_f2)


if __name__ == '__main__':
    unittest.main()




from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import create_feature_vectors
from wyrm.processing import swapaxes

class TestCreateFeatureVectors(unittest.TestCase):

    def setUp(self):
        # create epoched data with only 0s in class0, 1s in class1 and
        # 2s in class2
        cnt = np.ones((10, 3))
        epo = np.array([0*cnt, 1*cnt, 2*cnt])
        time = np.arange(10)
        channels = np.array(['ch1', 'ch2', 'ch3'])
        classes = np.arange(3)
        axes = ['class', 'time', 'channel']
        units = ['#', 'ms', '#']
        self.dat = Data(epo, [classes, time, channels], axes, units)

    def test_create_feature_vectors(self):
        """Create Feature Vectors."""
        dat = create_feature_vectors(self.dat)
        self.assertTrue(all(dat.data[0] == 0))
        self.assertTrue(all(dat.data[1] == 1))
        self.assertTrue(all(dat.data[2] == 2))
        self.assertEqual(dat.data.ndim, 2)
        self.assertEqual(len(dat.axes), 2)
        self.assertEqual(len(dat.names), 2)
        self.assertEqual(len(dat.units), 2)
        self.assertEqual(dat.names[-1], 'feature vector')
        self.assertEqual(dat.units[-1], 'dl')
        np.testing.assert_array_equal(dat.axes[-1], np.arange(dat.data.shape[-1]))

    def test_create_feature_vectors_copy(self):
        """create_feature_vectors must not modify argument."""
        cpy = self.dat.copy()
        create_feature_vectors(self.dat)
        self.assertEqual(cpy, self.dat)

    def test_create_feature_vectors_swapaxes(self):
        """create_feature_vectors must work with nonstandard classaxis."""
        # keep in mind that create_feature_vectors already swaps the
        # axes internally to move the classaxis to 0
        dat = create_feature_vectors(swapaxes(self.dat, 0, 2), classaxis=2)
        dat2 = create_feature_vectors(self.dat)
        self.assertEqual(dat, dat2)


if __name__ == '__main__':
    unittest.main()




from __future__ import division

import unittest

import numpy as np

from wyrm.types import Data
from wyrm.processing import remove_channels
from wyrm.processing import swapaxes

class TestRemoveChannels(unittest.TestCase):

    def setUp(self):
        raw = np.arange(20).reshape(4, 5)
        channels = ['ca1', 'ca2', 'cb1', 'cb2', 'cc1']
        time = np.arange(4)
        self.dat = Data(raw, [time, channels], ['time', 'channels'], ['ms', '#'])

    def test_remove_channels(self):
        """Removing channels with an array of regexes."""
        channels = self.dat.data.copy()
        self.dat = remove_channels(self.dat, ['ca.*', 'cc1'])
        np.testing.assert_array_equal(self.dat.axes[-1],
                                      np.array(['cb1', 'cb2']))
        np.testing.assert_array_equal(self.dat.data,
                                      channels[:, np.array([2, 3])])

    def test_remove_channels_swapaxis(self):
        """Remove channels works with non default chanaxis."""
        dat1 = remove_channels(swapaxes(self.dat, 0, 1), ['ca.*'], chanaxis=0)
        dat1 = swapaxes(dat1, 0, 1)
        dat2 = remove_channels(self.dat, ['ca.*'])
        self.assertEqual(dat1, dat2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python


import unittest
import os
import json
import struct

import numpy as np

from wyrm.io import load_mushu_data


class TestLoadMushuData(unittest.TestCase):

    def setUp(self):
        with open('foo.meta', 'w') as fh:
            json.dump({'Sampling Frequency' : 1,
                       'Channels' : ['ch1', 'ch2'],
                       'Amp': 'Dummy Amp'}, fh)
        with open('foo.marker', 'w') as fh:
            fh.write('0.0 Marker 0\n')
            fh.write('500.0 Marker 500\n')
            fh.write('666.666 Marker 666.666\n')
        with open('foo.eeg', 'wb') as fh:
            for i in (0, 1, 33, 66, .1, .2):
                fh.write(struct.pack('f', i))

    def tearDown(self):
        for fname in 'foo.meta', 'foo.eeg', 'foo.markers':
            try:
                os.remove(fname)
            except:
                pass

    def test_incomplete_fileset(self):
        """Must raise an error if not all files are available."""
        os.remove('foo.marker')
        with self.assertRaises(AssertionError):
            load_mushu_data('foo.meta')

    def test_loading(self):
        """Loads data correctly."""
        dat = load_mushu_data('foo.meta')
        self.assertEqual(dat.fs, 1)
        np.testing.assert_array_equal(dat.markers, [[0.0, 'Marker 0'], [500.0, 'Marker 500'], ['666.666', 'Marker 666.666']])
        np.testing.assert_array_equal(dat.axes[0], [0, 1000, 2000])
        np.testing.assert_array_equal(dat.axes[1], ['ch1', 'ch2'])
        self.assertEqual(dat.names, ['time', 'channels'])
        self.assertEqual(dat.units, ['ms', '#'])
        data = np.array([0, 1, 33, 66, .1, .2], np.float32).reshape(-1, 2)
        np.testing.assert_array_equal(dat.data, data)


if __name__ == '__main__':
    unittest.main()




from __future__ import division

import unittest

import random
import numpy as np

from wyrm.types import Data
from wyrm.processing import sort_channels, swapaxes, CHANNEL_10_20


class TestSortChannels(unittest.TestCase):

    def setUp(self):
        self.sorted_channels = np.array([name for name, pos in CHANNEL_10_20])
        channels = self.sorted_channels.copy()
        random.shuffle(channels)
        raw = np.random.random((5, 10, len(channels)))
        time = np.linspace(0, 1000, 10, endpoint=False)
        epochs = np.array([0, 1, 0, 1, 0])
        fs = 100
        marker = [[100, 'foo'], [200, 'bar']]
        self.dat = Data(raw, [epochs, time, channels], ['class', 'time', 'channels'], ['#', 'ms', '#'])
        self.dat.fs = fs
        self.dat.markers = marker

    def test_sort_channels(self):
        """sort_channels must sort correctly."""
        dat = sort_channels(self.dat)
        np.testing.assert_array_equal(dat.axes[-1], self.sorted_channels)

    def test_sort_channels_with_unknown_channel(self):
        """Unknown channels move to the back."""
        self.dat.axes[-1][7] = 'XX'
        dat = sort_channels(self.dat)
        self.assertEqual(dat.axes[-1][-1], 'XX')

    def test_sort_channels_swapaxis(self):
        """sort_channels must workt with nonstandard chanaxis."""
        sorted_ = sort_channels(swapaxes(self.dat, 1, -1), 1)
        sorted_ = swapaxes(sorted_, 1, -1)
        sorted2 = sort_channels(self.dat)
        self.assertEqual(sorted_, sorted2)

    def test_sort_channels_copy(self):
        """sort_channels must not modify argument."""
        cpy = self.dat.copy()
        sort_channels(self.dat)
        self.assertEqual(self.dat, cpy)

if __name__ == '__main__':
    unittest.main()
